from django.template.loader import render_to_string


class ColourRows:

    def __init__(self, datatable, colour_dict):
        self.colour_dict = colour_dict
        self.datatable = datatable

    def render(self):

        '''
        if 'rowColor' in self.table_options:
            if isinstance(self.table_options['rowColor'], list):
                for rc in self.table_options['rowColor']:
                    if 'colRef' in rc:
                        rc['column'] = self.all_refs().index(rc['colRef'])
            if 'colRef' in self.table_options['rowColor']:
                self.table_options['rowColor']['column'] = self.all_refs().index(self.table_options['rowColor']
                                                                                 ['colRef'])
        '''

        return render_to_string('datatables/plugins/colour_rows.html',
                                {'datatable': self.datatable, 'colour_dict': self.colour_dict})




def detect_device(request):
    if 'android' in request.META['HTTP_USER_AGENT'].lower() or 'iphone' in request.META['HTTP_USER_AGENT'].lower():
        return {'mobile':True}
    return {'mobile': False}



import json
from inspect import isclass
from typing import TypeVar, Dict
from django.db import models
from django.views.generic import TemplateView
from django.http import HttpResponse
from django.template.loader import render_to_string
from .detect_device import detect_device
from .columns import ColumnBase, DateColumn, ChoiceColumn, BooleanColumn, CallableColumn
from .model_def import DatatableModel
from .filters import DatatableFilter

KT = TypeVar('KT')
VT = TypeVar('VT')


class DatatableError(Exception):
    pass


class ColumnInitialisor:

    def __init__(self, start_model, path, field_prefix='', name_prefix='', **kwargs):
        self.start_model = start_model
        self.django_field = None
        self.setup = None
        self.model = None
        if type(path) == tuple:
            kwargs.update(path[1])
            path = path[0]
        self.kwargs = kwargs
        self.columns = []
        self.kwargs['model_path'] = field_prefix
        self.callable = False

        if isclass(path):
            self.setup = path
            self.kwargs['column_name'] = path.__name__
        elif isinstance(path, ColumnBase):
            self.setup = path
            if not hasattr(path, 'column_name'):
                self.kwargs['column_name'] = type(path).__name__
        elif isinstance(path, str):
            self.path, options = ColumnBase.extract_options(path)
            self.kwargs.update(options)
            self.model, self.django_field, self.setup = DatatableModel.get_setup_data(start_model, self.path)
            self.kwargs['column_name'] = path

            if '__' in self.path:
                split_path = self.path.split('__')
                self.field = split_path[-1]
                self.next_prefix = '__'.join(split_path[:-1]) + '__'
            else:
                self.field = self.path
                self.next_prefix = ''
            self.callable = callable(getattr(self.model, self.field, None))
        else:
            raise DatatableError('Unknown type in columns ' + str(path))

        if 'column_name' in kwargs and name_prefix:
            self.kwargs['column_name'] = f'{name_prefix}/{self.kwargs["column_name"]}'

    def get_columns(self):
        self.kwargs['model'] = self.model
        if isclass(self.setup):
            self.columns.append(self.setup(**self.kwargs))
        elif isinstance(self.setup, ColumnBase):
            if self.setup.initialised:
                self.columns.append(self.setup)
            else:
                self.columns.append(self.setup.get_class_instance(**self.kwargs))
        elif isinstance(self.setup, list):
            for c in self.setup:
                self.columns += ColumnInitialisor(start_model=self.start_model, path=c, field_prefix=self.next_prefix,
                                                  name_prefix=self.field, **self.kwargs).get_columns()
        elif self.callable:
            if self.setup is None:
                self.setup = {}
            self.columns.append(CallableColumn(field=self.field, **self.kwargs, **self.setup))
        else:
            self.kwargs['field'] = self.field
            if isinstance(self.setup, dict):
                self.kwargs.update(self.setup)
            if self.django_field:
                self.add_django_field_column()
            else:
                self.columns.append(ColumnBase(**self.kwargs))
        return self.columns

    def add_django_field_column(self):
        if 'title' not in self.kwargs:
            self.kwargs['title'] = self.django_field.verbose_name.title()
        field_type = type(self.django_field)
        if field_type in [models.DateField, models.DateTimeField]:
            self.columns.append(DateColumn(**self.kwargs))
        elif (field_type in [models.IntegerField, models.PositiveSmallIntegerField, models.PositiveIntegerField]
              and self.django_field.choices is not None and len(self.django_field.choices) > 0):
            self.columns.append(ChoiceColumn(choices=self.django_field.choices, **self.kwargs))
        elif field_type == models.BooleanField:
            self.columns.append(BooleanColumn(**self.kwargs))
        else:
            self.columns.append(ColumnBase(**self.kwargs))


class DatatableTable:

    def __init__(self, table_id, model=None, table_options=None, table_classes=None):
        self.columns = []
        self.table_id = table_id
        self.table_options: Dict[KT, VT] = {'pageLength': 100}
        if table_options:
            self.table_options.update(table_options)

        self.ajax_data = True
        self.model = model

        # django query attributes
        self.filter = {}
        self.exclude = {}
        self.distinct = []
        self.max_records = None

        # javascript datatable attributes
        self.order_by = []
        self.js_filter_list = []
        self.plugins = []
        self.omit_columns = []

        if table_classes:
            self.table_classes = table_classes
        else:
            self.table_classes = ['display', 'compact', 'smalltext', 'table-sm', 'table', 'w-100']

    def table_class(self):
        return ' '.join(self.table_classes)

    def column(self, column_name):
        return self.find_column(column_name)[0]

    def add_js_filters(self, name_or_template, column_ids, **kwargs):
        if type(column_ids) == str:
            column_ids = [column_ids]
        for c in column_ids:
            self.js_filter_list.append(DatatableFilter(name_or_template, self, column=self.find_column(c)[0], **kwargs))

    def js_filter(self, name_or_template, column_id, **kwargs):
        return DatatableFilter(name_or_template, self, column=self.find_column(column_id)[0], **kwargs)

    def add_plugin(self, plugin, *args, **kwargs):
        self.plugins.append(plugin(self, *args, **kwargs))

    def get_query(self, **_kwargs):
        annotations = {}
        for c in self.columns:
            if c.annotations:
                annotations.update(c.annotations)
        query = (self.model.objects.annotate(**annotations).exclude(**self.exclude).values(*self.fields())
                 .order_by(*self.order_by))
        if isinstance(self.filter, models.Q):
            query = query.filter(self.filter)
        else:
            query = query.filter(**self.filter)
        if self.distinct:
            query = query.distinct(*self.distinct)
        if self.max_records:
            query = query[:self.max_records]
        return query

    def sort(self, *columns):
        self.table_options['order'] = []
        for c in columns:
            sort_order = 'asc'
            if c[0] == '-':
                sort_order = 'desc'
                c = c[1:]
            self.table_options.setdefault('order', []).append([self.find_column(c)[1], sort_order])

    def row_color(self, column, *option_colors):
        self.table_options.setdefault('rowColor', []).append(
            {'colRef': column, 'values': {o[0]: o[1] for o in option_colors}})

    def add_column_options(self, column_names, options):
        if type(column_names) == str:
            column_names = [column_names]
        for c in self.columns:
            if c.column_name in column_names:
                c.setup_kwargs(options)

    def find_column(self, column_name, by_field=False):
        for n, c in enumerate(self.columns):
            if by_field and c.field == column_name:
                return c, n
            if c.column_name == column_name:
                return c, n
        raise DatatableError('Unable to find column ' + column_name)

    def add_columns(self, *columns):
        for c in columns:
            self.columns += ColumnInitialisor(self.model, c).get_columns()

    def fields(self):
        fields = []
        for c in self.columns:
            if c.field and 'calculated' not in c.options:
                if isinstance(c.field, (tuple, list)):
                    for f in c.field:
                        if f not in fields:
                            fields.append(f)
                else:
                    if c.field not in fields:
                        fields.append(c.field)
        return fields

    def all_names(self):
        return [c.column_name for c in self.columns]

    def all_titles(self):
        return [str(c.title) for c in self.columns]

    def render(self):
        rendered_strings = []
        for p in self.plugins:
            rendered_strings.append(p.render())
        rendered_strings.append(render_to_string('datatables/table.html', {'datatable': self}))
        return ''.join(rendered_strings)

    def setup_column_id(self):
        if 'column_id' not in self.table_options:
            try:
                self.table_options['column_id'] = self.find_column('id', True)[1]
                return self.columns[self.table_options['column_id']]
            except DatatableError:
                pass
        else:
            return self.columns[self.table_options['column_id']]

    def col_def_str(self):
        self.setup_column_id()
        options = dict(self.table_options)
        if not self.ajax_data:
            options['data'] = self.get_table_array(None, self.get_query())
        options['columnDefs'] = [dict({'targets': i, 'name': c.column_name}, **c.style())
                                 for i, c in enumerate(self.columns)]
        table_vars = {
            'field_ids': self.all_names(),
            'colOptions': [c.options for c in self.columns],
            'tableOptions': options,
        }
        col_def_str = json.dumps(table_vars, separators=(',', ':'))
        # legacy modifications to col_def_str
        # col_def_str = col_def_str.replace('"&', "")
        # col_def_str = col_def_str.replace('&"', "")
        return col_def_str

    def get_table_array(self, request, results):
        page_results = {}
        for c in self.columns:
            c.setup_results(request, page_results)
        return [[c.row_result(data_dict, page_results) for c in self.columns] for data_dict in results]

    def get_json(self, request, results):
        return_data = {'data': self.get_table_array(request, results)}
        return json.dumps(return_data, separators=(',', ':'), default=str)

    def refresh_row(self, request, row_id):
        id_column = self.setup_column_id()
        if id_column:
            self.filter[id_column.field] = int(row_id[1:])
            results = self.get_table_array(request, self.get_query())[0]
            commands = [{'command': 'refresh_row', 'row': row_id, 'data': results, 'table': self.table_id}]
            return HttpResponse(json.dumps(commands), content_type='application/json')

    def delete_row(self, _request, row_id):
        commands = [{'command': 'delete_row', 'row': row_id, 'table': self.table_id}]
        return HttpResponse(json.dumps(commands), content_type='application/json')


class DatatableView(TemplateView):
    model = None
    table_classes = None
    table_options = None

    def __init__(self, *args, **kwargs):
        super(DatatableView, self).__init__(*args, **kwargs)
        self.tables = {}
        self.add_tables()

        self.dispatch_context = None

    def add_table(self, table_id, **kwargs):
        self.tables[table_id] = DatatableTable(table_id, table_options=self.table_options,
                                               table_classes=self.table_classes,
                                               **kwargs)

    def add_tables(self):
        self.add_table(type(self).__name__.lower(), model=self.model)

    def setup_tables(self, table_id=None):
        for t_id, table in self.tables.items():
            if not table_id or t_id == table_id:
                if t_id == type(self).__name__.lower():
                    self.setup_table(table)
                else:
                    getattr(self, 'setup_' + t_id)(table)

    def dispatch(self, request, *args, **kwargs):
        self.dispatch_context = detect_device(request)
        return super(DatatableView, self).dispatch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        self.setup_tables()
        return super().get(request, *args, **kwargs)

    # def render_to_response(self, context, **response_kwargs):
    #    if self.template_name is None:
    #        return HttpResponse(self.table.javascript_setup())
    #    return super(DatatableView, self).render_to_response(context, **response_kwargs)

    # @staticmethod
    # def graph_data(**kwargs):
    #    return None
    @staticmethod
    def setup_table(table):
        pass

    @staticmethod
    def locked():
        return False

    @staticmethod
    def get_table_query(table, **kwargs):
        return table.get_query(**kwargs)

    def post(self, request, *args, **kwargs):

        if 'datatable-data' in request.GET:
            table = self.tables[request.POST['table_id']]
            self.setup_tables(table_id=table.table_id)
            results = self.get_table_query(table, **kwargs)
            if self.locked():
                secure_fields = [c.field for c in table.columns if c.options.get('secure')]
                if secure_fields:
                    for r in results:
                        for f in secure_fields:
                            data = r.get(f)
                            if data:
                                r[f] = '<i class="fas fa-key"></i>'
            return HttpResponse(table.get_json(request, results), content_type='application/json')

        for t in ['row', 'column']:
            if 'datatable-' + t in request.GET:
                table = self.tables[request.POST['table_id']]
                self.setup_tables(table_id=table.table_id)
                if hasattr(self, t + '_' + request.POST['command']):
                    column_values = json.loads(request.POST[t])
                    extra_data = {k: request.POST[k] for k in request.POST if k != t}
                    return getattr(self, t + '_' + request.POST['command'])(request, column_values, table, extra_data)
        return None

    def sent_column(self, column_values, extra_data):
        pass
    """
    def get(self, request, *args, **kwargs):
        if request.GET.get('json', None) is not None:
            '''
            # ***************************************************
            # Uncomment for benchmarking SQL calls for datatables
            # ***************************************************
            a = getJson(self.table, self.post_table_json(**kwargs))
            context = self.get_context_data(**kwargs)
            self.template_name = 'blank.html'
            return self.render_to_response(context)
            '''
            return self.table_json(**kwargs)
        else:
            context = self.get_context_data(**kwargs)
            return self.render_to_response(context)
    """
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        if len(self.tables) == 1:
            context['datatable'] = self.tables[list(self.tables.keys())[0]]
        else:
            context['datatables'] = self.tables
        context.update(self.add_to_context(**kwargs))
        return context

    def add_to_context(self, **kwargs):
        return {}


simple_table = {
    'dom': 't',
    'no_col_search': True,
    'nofooter': True,
    'pageLength': 400,
}



from django.urls import reverse
DUMMY_ID = 999999


def row_button(command, button_text, *, function='Html', button_classes='btn btn-sm', **kwargs):
    rb = {
        'html': (f'<button data-command="{command}" onclick="django_datatables.b_r(this)" '
                 f'class="{button_classes}">{button_text}</button>'),
        'function': function,
    }
    rb.update(kwargs)
    return rb


def render_replace(*, var='%1%', **kwargs):
    return dict(var=var, function='Replace', **kwargs)


def get_url(url_name):
    if type(url_name) == tuple:
        return reverse(url_name[0], args=[*url_name[1:]])
    else:
        if url_name.find(str(DUMMY_ID)) == -1:
            return reverse(url_name, args=[DUMMY_ID])
        return url_name


def row_link(url_name, column_id):
    return [render_replace(column=column_id, html=get_url(url_name), var=str(DUMMY_ID))]



import string
import random
from django.template.loader import render_to_string


class DatatableFilter:

    template_library = {
        'pivot': 'datatables/filter_blocks/pivot_filter.html',
        'tag': 'datatables/filter_blocks/tag_filter.html',
        'totals': 'datatables/filter_blocks/pivot_totals.html',
        'select2': 'datatables/filter_blocks/select2_filter.html',
        'date': 'datatables/filter_blocks/date_filter.html',
        'expand': 'datatables/filter_blocks/row_tree.html',
    }

    def __init__(self, name_or_template, datatable, column=None, collapsed=False, filter_title=None, html_id=None,
                 **kwargs):
        self.template = ""
        self.column = column
        self.datatable = datatable
        self.collapsed = collapsed
        self.kwargs = kwargs
        if html_id:
            self.html_id = html_id
        else:
            letters = string.ascii_lowercase
            self.html_id = ''.join(random.choice(letters) for _i in range(10))

        if filter_title:
            self.filter_title = filter_title
        elif column:
            self.filter_title = column.title
        else:
            self.filter_title = ''

        if name_or_template in self.template_library:
            self.template = self.template_library[name_or_template]
        else:
            self.template = name_or_template

    def render(self):
        context = {
            'html_id': self.html_id,
            'filter_title': self.filter_title,
            'table': self.datatable,
            'collapsed': self.collapsed,
            'kwargs': self.kwargs
        }
        if self.column:
            context['column_no'] = self.datatable.find_column(self.column.column_name)[1]

        return render_to_string(self.template, context)



from django import template
from django.utils.safestring import mark_safe
from django_datatables.filters import DatatableFilter
register = template.Library()


@register.simple_tag
def filter_template(name_or_template, datatable, **kwargs):
    return DatatableFilter(name_or_template, datatable, **kwargs).render()


@register.inclusion_tag('datatables/filter_blocks/table_search.html')
def table_search(table):
    return {'table': table}


@register.inclusion_tag('datatables/filter_blocks/filter_control.html')
def filter_control(table):
    return {'datatable': table}


@register.simple_tag
def filter_group(datatable):
    filters = []
    for f in datatable.js_filter_list:
        filters.append(f.render())
    return mark_safe(''.join(filters))


@register.inclusion_tag('datatables/no_of_rows.html')
def no_of_rows(table):
    return {'datatable': table}


@register.inclusion_tag('datatables/clear_table.html')
def clear_table(table):
    return {'datatable': table}


@register.inclusion_tag('datatables/no_of_results.html')
def no_of_results(table):
    return {'datatable': table}



import re
import inspect
from types import MethodType
from typing import TypeVar, Dict
from .helpers import get_url, render_replace
KT = TypeVar('KT')
VT = TypeVar('VT')


class ColumnNameError(Exception):
    pass


class DatatableColumnError(Exception):
    pass


class ColumnBase:

    @staticmethod
    def merge_kwargs_locals(local_vars):
        kwargs = local_vars.pop('kwargs', {})
        kwargs.update(local_vars)
        del kwargs['self']
        kwargs.pop('__class__', None)
        return kwargs

    def initialise(self, local_vars):
        if not hasattr(self, 'initialised'):
            self.kwargs = self.merge_kwargs_locals(local_vars)
            # noinspection PyAttributeOutsideInit
            self.initialised = 'column_name' in self.kwargs or hasattr(self, 'column_name')
        return self.initialised

    def get_class_instance(self, column_name, **kwargs):
        kwargs.update(self.kwargs)
        return self.__class__(column_name=column_name, class_holder=False, **kwargs)

    def extract_args(self):
        args = self.column_name.split('|')
        return_args = []
        if len(args) > 1:
            self.column_name = args[0]
            for a in args[1:]:
                split_arg = a.split('=')
                if len(split_arg) == 2:
                    self.kwargs[split_arg[0]] = split_arg[1]
                else:
                    return_args.append(a)
        return return_args

    def __init__(self, field=None, **kwargs):
        if not self.initialise(locals()):
            return
        # Remove  ._$ characters from beginning of name and set appropriate options
        self._column_name = None
        self.options: Dict[KT, VT] = {}
        self.column_defs = {}
        self.column_name, options = self.extract_options(kwargs.get('column_name', ''))
        self.options: Dict[KT, VT] = options
        self.model_path = kwargs.pop('model_path', None)
        if not self.model_path:
            self.model_path = self.get_model_path(self.column_name)
        self.model = kwargs.pop('model', None)
        if kwargs.pop('column_name', None) == field:
            self.field = self.column_name
        else:
            self.field = field
        self.args = self.extract_args()
        self.title = self.title_from_name(self.column_name)
        self.column_type = 0
        self._annotations = None
        self.additional_columns = []
        self.kwargs = kwargs
        self.replace_list = []
        if not hasattr(self, 'row_result'):
            self.row_result = MethodType(self.__row_result, self)
        self.setup_kwargs(kwargs)

    @property
    def annotations(self):
        return self._annotations

    @annotations.setter
    def annotations(self, value):
        for f in value:
            if self.field is None:
                self._field = f
            elif type(self.field) == str:
                if f != self.field:
                    self._field = [self.field, f]
            else:
                if f not in self.field:
                    self._field.append(f)
        self._annotations = value

    @property
    def column_name(self):
        return self._column_name

    @column_name.setter
    def column_name(self, value):
        if not re.match('^[A-Za-z0-9._/]+$', value):
            raise ColumnNameError('Invalid column_name: ' + value)
        self._column_name = value

    @property
    def field(self):
        return self._field

    @field.setter
    def field(self, fields):
        if fields is None:
            self._field = None
        elif isinstance(fields, (list, tuple)):
            self.options['field_array'] = True
            self._field = [self.model_path + o for o in fields]
            if not hasattr(self, 'row_result') or type(self.row_result.__self__) == ColumnBase:
                self.row_result = MethodType(self.__list_row_result, self)
        else:
            self._field = self.model_path + fields

    def field_append(self, field):
        if not isinstance(self._field, (list, tuple)):
            self._field = [self._field, self.model_path + field]
        else:
            self._field = list(self._field).append(self.model_path + field)

    @staticmethod
    def get_model_path(field):
        if '__' in field:
            return '__'.join(field.split('__')[:-1]) + '__'
        else:
            return ''

    @staticmethod
    def extract_options(field):
        attributes = re.search('^[._$]+', field)
        options = {}
        if attributes:
            field = field[attributes.end():]
            if '_' in attributes.group():
                options['calculated'] = True
            if '.' in attributes.group():
                options['hidden'] = True
            if '$' in attributes.group():
                options['secure'] = True
        return field, options

    def setup_results(self, request, all_results):
        return

    @staticmethod
    def __list_row_result(self, data_dict, _page_results):
        return [data_dict.get(f) for f in self.field]

    @staticmethod
    def __row_result(self, data_dict, _page_results):
        if self.options.get('choices'):
            return self.options['choices'].get(data_dict.get(self.field))
        return data_dict.get(self.field)

    def style(self):
        col_def_str = self.column_defs
        if self.options.get('hidden'):
            col_def_str['visible'] = False
            col_def_str['searchable'] = False
        return col_def_str

    def replace_list_append(self, replace, column, in_column=None, values=None):
        replace_options = {'column': column, 'var': replace}
        if in_column:
            replace_options.update({'in': in_column, 'values': values})
        self.replace_list.append(replace_options)

    @staticmethod
    def title_from_name(name):
        if type(name) == str and len(name) > 0:
            field_no_path = name.split('__')[-1]
            if field_no_path.find('_') > 0:
                return field_no_path.replace('_', ' ').title()
            else:
                title = field_no_path[0]
                for letter in field_no_path[1:]:
                    if letter.isupper():
                        title += ' '
                    title += letter
                return title

    def setup_kwargs(self, kwargs):

        for a, value in kwargs.items():
            if a in ['field', 'column_name']:
                continue
            if a == 'row_result':
                self.row_result = MethodType(value, self)
            elif a == 'width':
                self.column_defs['width'] = value
            elif hasattr(self, a):
                setattr(self, a, value)
            else:
                self.options[a] = value
        return self


class DatatableColumn(ColumnBase):
    def __init__(self, **kwargs):
        if not self.initialise(locals()):
            return
        super().__init__(**kwargs)
        self.col_setup()

    def col_setup(self):
        pass


class LambdaColumn(ColumnBase):

    def __init__(self, *, lambda_function, **kwargs):
        if not self.initialise(locals()):
            return
        super().__init__(**kwargs)
        self.lambda_function = lambda_function

    def row_result(self, data_dict, _page_results):
        return self.lambda_function(data_dict.get(self.field))


class TextFieldColumn(ColumnBase):

    def row_result(self, data_dict, _page_results):
        result = data_dict.get(self.field, '')
        if len(result) > self.kwargs.get('max_chars'):
            result = result[:self.kwargs.get('max_chars')] + '...'
        return result.replace('\n', '<br>')

    def __init__(self, max_chars=8000, **kwargs):
        if not self.initialise(locals()):
            return
        super().__init__(**self.kwargs)
        if len(self.args) > 0:
            self.kwargs['max_chars'] = int(self.args[0])


class ColumnLink(ColumnBase):

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, url_name):
        self._url = get_url(url_name)

    def __init__(self, *, url_name, link_ref_column='id', link_html='%1%', var='%1%', **kwargs):
        if not self.initialise(locals()):
            return
        super().__init__(**self.kwargs)
        self.url = url_name

        if isinstance(self.field, (list, tuple)):
            self.options['render'] = [render_replace(column=self.column_name + ':0',
                                                     html=f'<a href="{self.url}">%1%</a>',
                                                     var='999999'),
                                      render_replace(column=self.column_name + ':1'),
                                      ]
        elif var not in link_html:
            self.options['render'] = [render_replace(column=link_ref_column,
                                                     html=f'<a href="{self.url}">{link_html}</a>',
                                                     var='999999')]
        else:
            self.options['render'] = [render_replace(column=self.column_name,
                                                     html=f'<a href="{self.url}">{link_html}</a>',
                                                     var=var), render_replace(column=link_ref_column, var='999999')]


class ManyToManyColumn(DatatableColumn):

    def setup_results(self, request, all_results):
        if self.reverse:
            tags = self.related_model.objects.values_list(self.field_id, 'pk')
        else:
            tags = self.model.objects.values_list('pk', self.field_id)
        tags = tags.filter(**{self.field_id + '__isnull': False}).distinct()
        tag_dict = {}
        for t in tags:
            tag_dict.setdefault(t[0], []).append(t[1])
        all_results['m2m' + self.column_name] = tag_dict

    def row_result(self, data_dict, page_results):
        return page_results['m2m' + self.column_name].get(data_dict['id'], self.blank)

    def __init__(self, *,  html=' %1% ', **kwargs):
        if not self.initialise(locals()):
            return
        super().__init__(**self.kwargs)
        fields = self.field.split('__')
        if not inspect.isclass(self.model):
            raise DatatableColumnError('ManyToManyColumn must have model set')
        connecting_field = self.model._meta.get_field(fields[-2])
        self.related_model = connecting_field.related_model
        if hasattr(connecting_field, 'field'):
            self.field_id = connecting_field.field.attname + '__pk'
            self.reverse = True
        else:
            self.field_id = fields[-2] + '__pk'
            self.reverse = False
        self.field = None
        self.options['lookup'] = kwargs.pop('lookup', list(self.related_model.objects.values_list('pk', fields[-1])))
        if 'blank' in kwargs:
            self.options['lookup'].append((-1, kwargs.pop('blank')))
            self.blank = [-1]
        else:
            self.blank = []
        self.options['render'] = [{'var': '%1%', 'html': html, 'function': 'ReplaceLookup'}]
        self.setup_kwargs(kwargs)


class DateColumn(ColumnBase):

    def row_result(self, data, _page_data):
        try:
            date = data[self.field].strftime('%d/%m/%Y')
            return date
        except AttributeError:
            return ""


class DateTimeColumn(ColumnBase):

    def row_result(self, data, _page_data):
        try:
            date = data[self.field].strftime('%d/%m/%Y')
            time_str = data[self.field].strftime('%H:%M')
            return date + ' ' + time_str
        except AttributeError:
            return ""


class ChoiceColumn(ColumnBase):

    def setup_kwargs(self, kwarg_dict):
        choices = kwarg_dict.pop('choices')
        self.options = {c[0]: c[1] for c in choices}
        super().setup_kwargs(kwarg_dict)

    def row_result(self, data, _page_data):
        return self.options.get(data[self.field], '')


class CurrencyPenceColumn(ColumnBase):

    def row_result(self, data, _page_data):
        try:
            return '{:.2f}'.format(data[self.field] / 100.0)
        except KeyError:
            return


class CurrencyColumn(ColumnBase):

    def row_result(self, data, _page_data):
        try:
            return '{:.2f}'.format(data[self.field])
        except KeyError:
            return


class BooleanColumn(ColumnBase):

    choices = ['true', 'false']

    def row_result(self, data, _page_data):
        try:
            if data[self.field]:
                return self.choices[0]
            else:
                return self.choices[1]
        except KeyError:
            return


class GroupedColumn(DatatableColumn):

    page_key = ''

    @staticmethod
    def initial_column_data(_request):
        return ''

    def setup_results(self, request, all_results):
        all_results[self.page_key] = self.initial_column_data(request)

    def row_result(self, data, page_data):
        return page_data[self.page_key].get(data[self.field])


class CallableColumn(DatatableColumn):

    def row_result(self, data_dict, _page_results):
        fake_object = type('fake', (), {k[len(self.model_path):]: v
                                        for k, v in data_dict.items()
                                        if k.startswith(self.model_path)})
        return self.object_function(fake_object)

    def setup_kwargs(self, kwargs):
        super().setup_kwargs(kwargs)
        self.object_function = getattr(self.model, self.field[len(self.model_path):])
        self.field = kwargs.get('parameters')



from inspect import signature, isclass
from django.core.exceptions import FieldDoesNotExist


class DatatableModel:

    @staticmethod
    def column_info(field_name, **kwargs):
        return

    @staticmethod
    def extract_path(field):
        if '__' in field:
            return field[:field.rfind('__') + 2]
        return ''

    @classmethod
    def get_column_info(cls, model, field_name, **kwargs):
        field_id = field_name.split('__')[-1]
        try:
            config = getattr(model.Datatable, field_id)
            if callable(config):
                if isclass(config):
                    column_setup = config(field_name).column()
                elif len(signature(config).parameters) == 0:
                    column_setup = config()
                else:
                    column_setup = config(field_name)
            else:
                column_setup = config
        except AttributeError:
            column_setup = model.Datatable.column_info(field_id, **kwargs)

        if type(column_setup) == dict:
            column_setup['column_name'] = field_id
        return column_setup

    def table_column_definitions(self):
        pass

    def __init__(self):
        pass

    @staticmethod
    def get_setup_data(model, field):
        """
        Recursive routine to follow foreign keys to get base model for field and check if DataTable class exists
        :param model:
        :param field:
        :return:
        """
        split_field = field.split('__')
        setup = None
        if len(split_field) == 1:
            if hasattr(model, 'Datatable'):
                try:
                    setup = getattr(model.Datatable, field.split('|')[0])
                except AttributeError:
                    setup = model.Datatable.column_info(field)
            try:
                field = model._meta.get_field(field)
            except FieldDoesNotExist:
                field = None
            return model, field, setup
        else:
            next_field = model._meta.get_field(split_field[0])
            if next_field.many_to_one:
                next_model = next_field.foreign_related_fields[0].model
            elif next_field.many_to_many or next_field.one_to_one or next_field.one_to_many:
                next_model = next_field.related_model
            return DatatableModel.get_setup_data(next_model, '__'.join(split_field[1:]))